# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import Callable, List

import pytest
import sciline
import scipp as sc

import esssans as sans
from esssans.types import (
    BackgroundRun,
    BackgroundSubtractedIofQ,
    BeamCenter,
    CorrectForGravity,
    DirectBeam,
    DirectBeamFilename,
    DirectRun,
    Filename,
    Incident,
    IofQ,
    NeXusMonitorName,
    NonBackgroundWavelengthRange,
    QBins,
    RawData,
    SampleRun,
    SolidAngle,
    Transmission,
    UncertaintyBroadcastMode,
    WavelengthBands,
    WavelengthBins,
    WavelengthMask,
)


def make_params() -> dict:
    params = {}
    params[NeXusMonitorName[Incident]] = 'monitor2'
    params[NeXusMonitorName[Transmission]] = 'monitor4'
    band = sc.linspace('wavelength', 2.0, 16.0, num=2, unit='angstrom')
    params[WavelengthBands] = band
    params[WavelengthBins] = sc.linspace(
        'wavelength', start=band[0], stop=band[-1], num=141
    )
    params[WavelengthMask] = sc.DataArray(
        data=sc.array(dims=['wavelength'], values=[True]),
        coords={
            'wavelength': sc.array(
                dims=['wavelength'], values=[2.21, 2.59], unit='angstrom'
            )
        },
    )

    params[QBins] = sc.linspace(
        dim='Q', start=0.01, stop=0.55, num=141, unit='1/angstrom'
    )
    params[Filename[BackgroundRun]] = 'SANS2D00063159.hdf5'
    params[Filename[SampleRun]] = 'SANS2D00063114.hdf5'
    params[Filename[DirectRun]] = 'SANS2D00063091.hdf5'
    params[DirectBeamFilename] = 'DIRECT_SANS2D_REAR_34327_4m_8mm_16Feb16.hdf5'
    params[NonBackgroundWavelengthRange] = sc.array(
        dims=['wavelength'], values=[0.7, 17.1], unit='angstrom'
    )
    params[CorrectForGravity] = True
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound
    return params


def sans2d_providers():
    return sans.providers + sans.sans2d.providers


def test_can_create_pipeline():
    sciline.Pipeline(sans2d_providers(), params=make_params())


@pytest.mark.parametrize(
    'uncertainties',
    [UncertaintyBroadcastMode.drop, UncertaintyBroadcastMode.upper_bound],
)
def test_pipeline_can_compute_background_subtracted_IofQ(uncertainties):
    params = make_params()
    params[UncertaintyBroadcastMode] = uncertainties
    pipeline = sciline.Pipeline(sans2d_providers(), params=params)
    result = pipeline.compute(BackgroundSubtractedIofQ)
    assert result.dims == ('Q',)


def test_workflow_is_deterministic():
    params = make_params()
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.drop
    pipeline = sciline.Pipeline(sans2d_providers(), params=params)
    # This is Sciline's default scheduler, but we want to be explicit here
    scheduler = sciline.scheduler.DaskScheduler()
    graph = pipeline.get(IofQ[SampleRun], scheduler=scheduler)
    reference = graph.compute().data
    result = graph.compute().data
    assert sc.identical(sc.values(result), sc.values(reference))


def test_pipeline_raisesVariancesError_if_normalization_errors_not_dropped():
    params = make_params()
    del params[NonBackgroundWavelengthRange]  # Make sure we raise in iofq_denominator
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.fail
    pipeline = sciline.Pipeline(sans2d_providers(), params=params)
    with pytest.raises(sc.VariancesError):
        pipeline.compute(BackgroundSubtractedIofQ)


def test_uncertainty_broadcast_mode_drop_yields_smaller_variances():
    params = make_params()
    # Errors with the full range have some NaNs or infs
    params[QBins] = sc.linspace(
        dim='Q', start=0.01, stop=0.5, num=141, unit='1/angstrom'
    )
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.drop
    pipeline = sciline.Pipeline(sans2d_providers(), params=params)
    drop = pipeline.compute(IofQ[SampleRun]).hist().data
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound
    pipeline = sciline.Pipeline(sans2d_providers(), params=params)
    upper_bound = pipeline.compute(IofQ[SampleRun]).data
    assert sc.all(sc.variances(drop) < sc.variances(upper_bound)).value


def test_pipeline_can_visualize_background_subtracted_IofQ():
    pipeline = sciline.Pipeline(sans2d_providers(), params=make_params())
    pipeline.visualize(BackgroundSubtractedIofQ)


def test_pipeline_can_compute_intermediate_results():
    pipeline = sciline.Pipeline(sans2d_providers(), params=make_params())
    result = pipeline.compute(SolidAngle[SampleRun])
    assert result.dims == ('spectrum',)


# TODO See scipp/sciline#57 for plans on a builtin way to do this
def as_dict(funcs: List[Callable[..., type]]) -> dict:
    from typing import get_type_hints

    return dict(zip([get_type_hints(func)['return'] for func in funcs], funcs))


def pixel_dependent_direct_beam(
    filename: DirectBeamFilename, shape: RawData[SampleRun]
) -> DirectBeam:
    direct_beam = sans.sans2d.pooch_load_direct_beam(filename)
    sizes = {'spectrum': shape.sizes['spectrum'], **direct_beam.sizes}
    return DirectBeam(direct_beam.broadcast(sizes=sizes).copy())


@pytest.mark.parametrize(
    'uncertainties',
    [UncertaintyBroadcastMode.drop, UncertaintyBroadcastMode.upper_bound],
)
def test_pixel_dependent_direct_beam_is_supported(uncertainties):
    params = make_params()
    params[UncertaintyBroadcastMode] = uncertainties
    providers = as_dict(sans2d_providers())
    providers[DirectBeam] = pixel_dependent_direct_beam
    pipeline = sciline.Pipeline(providers.values(), params=params)
    result = pipeline.compute(BackgroundSubtractedIofQ)
    assert result.dims == ('Q',)


def test_beam_center_from_center_of_mass_is_close_to_verified_result():
    params = make_params()
    providers = sans2d_providers()
    pipeline = sciline.Pipeline(providers, params=params)
    center = pipeline.compute(BeamCenter)
    # This is the result we got with the pre-sciline implementation, using the full IofQ
    # calculation. The difference is about 5 mm in X or Y, probably due to a bias
    # introduced by the sample holder, which the center-of-mass approach cannot ignore.
    center_pre_sciline_raw_solid_angle = sc.vector([0.0945643, -0.082074, 0], unit='m')
    assert sc.allclose(
        center, center_pre_sciline_raw_solid_angle, atol=sc.scalar(5e-3, unit='m')
    )


def test_beam_center_finder_without_direct_beam_reproduces_verified_result():
    params = make_params()
    params[sans.beam_center_finder.BeamCenterFinderQBins] = sc.linspace(
        'Q', 0.02, 0.3, 71, unit='1/angstrom'
    )
    del params[DirectBeamFilename]
    providers = sans2d_providers()
    providers.remove(sans.beam_center_finder.beam_center_from_center_of_mass)
    providers.append(sans.beam_center_finder.beam_center_from_iofq)
    pipeline = sciline.Pipeline(providers, params=params)
    center = pipeline.compute(BeamCenter)
    # This is the result we got with the pre-sciline implementation
    # The difference is that the reference result computed the solid angle only once,
    # before applying any detector positions shifts.
    center_pre_sciline_raw_solid_angle = sc.vector([0.0945643, -0.082074, 0], unit='m')
    assert sc.allclose(
        center, center_pre_sciline_raw_solid_angle, atol=sc.scalar(4e-3, unit='m')
    )


def test_beam_center_finder_works_with_direct_beam():
    params = make_params()
    params[sans.beam_center_finder.BeamCenterFinderQBins] = sc.linspace(
        'Q', 0.02, 0.3, 71, unit='1/angstrom'
    )
    providers = sans2d_providers()
    providers.remove(sans.beam_center_finder.beam_center_from_center_of_mass)
    providers.append(sans.beam_center_finder.beam_center_from_iofq)
    pipeline = sciline.Pipeline(providers, params=params)
    center = pipeline.compute(BeamCenter)  # (0.0951122, -0.079375, 0)
    center_no_direct_beam = sc.vector([0.0945643, -0.082074, 0], unit='m')

    assert sc.allclose(center, center_no_direct_beam, atol=sc.scalar(1e-2, unit='m'))


def test_beam_center_finder_works_with_pixel_dependent_direct_beam():
    params = make_params()
    params[sans.beam_center_finder.BeamCenterFinderQBins] = sc.linspace(
        'Q', 0.02, 0.3, 71, unit='1/angstrom'
    )
    providers = sans2d_providers()
    providers.remove(sans.beam_center_finder.beam_center_from_center_of_mass)
    providers.append(sans.beam_center_finder.beam_center_from_iofq)
    pipeline = sciline.Pipeline(providers, params=params)
    center_pixel_independent_direct_beam = pipeline.compute(BeamCenter)

    direct_beam = (
        pipeline.compute(DirectBeam)
        .broadcast(sizes={'spectrum': 61440, 'wavelength': 175})
        .copy()
    )

    del params[DirectBeamFilename]
    params[DirectBeam] = direct_beam
    # Hack to remove direct-beam provider, until Sciline API improved
    providers = sans.providers + sans.sans2d.providers[1:]
    providers.remove(sans.beam_center_finder.beam_center_from_center_of_mass)
    providers.append(sans.beam_center_finder.beam_center_from_iofq)
    pipeline = sciline.Pipeline(providers, params=params)

    center = pipeline.compute(BeamCenter)
    assert sc.identical(center, center_pixel_independent_direct_beam)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen
"""
Utilities for logging.

All code in the ess package should log auxiliary information
through this module.
Loggers should be obtained through :py:func:`ess.logging.get_logger`
and pass a name that reflects the current context.
E.g., in the loki package, pass ``'loki'`` (all lowercase).

Logging can be configured using :py:func:`ess.logging.configure` or
:py:func:`ess.logging.configure_workflow`.
Use the latter at the beginning of workflow notebooks.
"""

import functools
import inspect
import logging
import logging.config
from copy import copy
from os import PathLike
from typing import Any, Callable, List, Optional, Sequence, Union

import scipp as sc
import scippneutron as scn
from scipp.utils import running_in_jupyter


def get_logger(subname: Optional[str] = None) -> logging.Logger:
    """Return one of ess's loggers.

    Parameters
    ----------
    subname:
        Name of an instrument, technique, or workflow.
        If given, return the logger with the given name
        as a child of the ess logger.
        Otherwise, return the general ess logger.

    Returns
    -------
    :
        The requested logger.
    """
    name = 'scipp.ess' + ('.' + subname if subname else '')
    return logging.getLogger(name)


def log_call(
    *, instrument: str, message: str = None, level: Union[int, str] = logging.INFO
):
    """
    Decorator that logs a message every time the function is called.
    """
    level = logging.getLevelName(level) if isinstance(level, str) else level

    def deco(f: Callable):
        @functools.wraps(f)
        def impl(*args, **kwargs):
            if message is not None:
                get_logger(instrument).log(level, message)
            else:
                get_logger(instrument).log(level, 'Calling %s', _function_name(f))
            return f(*args, **kwargs)

        return impl

    return deco


class Formatter(logging.Formatter):
    """
    Logging formatter that indents messages and optionally shows threading information.
    """

    def __init__(self, show_thread: bool, show_process: bool):
        """
        Initialize the formatter.

        The formatting is mostly fixed.
        Only printing of thread and processor names can be toggled using the
        corresponding arguments.
        Times are always printed in ISO 8601 format.
        """
        fmt_proc = '%(processName)s' if show_process else ''
        fmt_thread = '%(threadName)s' if show_thread else ''
        if show_process:
            fmt_proc_thread = fmt_proc
            if show_thread:
                fmt_proc_thread += ',' + fmt_thread
        elif show_thread:
            fmt_proc_thread = fmt_thread
        else:
            fmt_proc_thread = ''
        fmt_pre = '[%(asctime)s] %(levelname)-8s '
        fmt_post = '<%(name)s> : %(message)s'
        fmt = (
            fmt_pre
            + ('{' + fmt_proc_thread + '} ' if fmt_proc_thread else '')
            + fmt_post
        )
        super().__init__(fmt, datefmt='%Y-%m-%dT%H:%M:%S%z')

    def format(self, record: logging.LogRecord) -> str:
        record = copy(record)
        record.msg = '\n    ' + record.msg.replace('\n', '\n    ')
        return super().format(record)


def default_loggers_to_configure() -> List[logging.Logger]:
    """
    Return a list of all loggers that get configured by ess by default.
    """
    import pooch

    return [
        sc.get_logger(),
        logging.getLogger('Mantid'),
        pooch.get_logger(),
    ]


def configure(
    *,
    filename: Optional[Union[str, PathLike]] = 'scipp.ess.log',
    file_level: Union[str, int] = logging.INFO,
    stream_level: Union[str, int] = logging.WARNING,
    widget_level: Union[str, int] = logging.INFO,
    show_thread: bool = False,
    show_process: bool = False,
    loggers: Optional[Sequence[Union[str, logging.Logger]]] = None,
):
    """Set up logging for the ess package.

    This function is meant as a helper for application (or notebook) developers.
    It configures the loggers of ess, scippneutron, scipp, and some
    third party packages.
    *Calling it from a library should be avoided*
    because it can mess up a user's setup.

    Up to 3 handlers are configured:

      - *File Handler* Writes files to a file with given name.
        Can be disabled using the `filename` argument.
      - *Stream Handler* Writes to `sys.stderr`.
      - *Widget Handler* Writes to a :py:class:`scipp.logging.LogWidget`
        if Python is running in a Jupyter notebook.

    Parameters
    ----------
    filename:
        Name of the log file. Overwrites existing files.
        Setting this to `None` disables logging to file.
    file_level:
        Log level for the file handler.
    stream_level:
        Log level for the stream handler.
    widget_level:
        Log level for the widget handler.
    show_thread:
        If `True`, log messages include the name of the thread
        the message originates from.
    show_process:
        If `True`, log messages include the name of the process
        the message originates from.
    loggers:
        Collection of loggers or names of loggers to configure.
        If not given, uses :py:func:`default_loggers_to_configure`.

    See Also
    --------
    ess.logging.configure_workflow:
        Configure logging and do some additional setup for a reduction workflow.
    """
    if configure.is_configured:
        get_logger().warning(
            'Called `logging.configure` but logging is already configured'
        )
        return

    handlers = _make_handlers(
        filename, file_level, stream_level, widget_level, show_thread, show_process
    )
    base_level = _base_level([file_level, stream_level, widget_level])
    loggers = {
        logging.getLogger(logger) if isinstance(logger, str) else logger
        for logger in (default_loggers_to_configure() if loggers is None else loggers)
    }
    for logger in loggers:
        _configure_logger(logger, handlers, base_level)
    if any(logger.name == 'Mantid' for logger in loggers):
        _configure_mantid_logging('notice')

    configure.is_configured = True


configure.is_configured = False


def configure_workflow(
    workflow_name: Optional[str] = None, *, display: Optional[bool] = None, **kwargs
) -> logging.Logger:
    """Configure logging for a reduction workflow.

    Configures loggers, logs a greeting message, sets up a logger for a workflow,
    and optionally creates and displays a log widget.

    Parameters
    ----------
    workflow_name:
        Used as the name of the returned logger.
    display:
        If `True`, show a :py:class:`scipp.logging.LogWidget`
        in the outputs of the current cell.
        Defaults to `True` in Jupyter and `False` otherwise.
    kwargs:
        Forwarded to :py:func:`ess.logging.configure`.
        Refer to that function for details.

    Returns
    -------
    :
        A logger for use in the workflow.

    See Also
    --------
    ess.logging.configure:
        General logging setup.
    """
    configure(**kwargs)
    greet()
    if (display is None and running_in_jupyter()) or display:
        sc.display_logs()
    return get_logger(workflow_name)


def greet():
    """Log a message showing the versions of important packages."""
    # Import here so we don't import from a partially built package.
    from . import __version__

    msg = f'''Software Versions:
  ess: {__version__} (https://scipp.github.io/ess)
  scippneutron: {scn.__version__} (https://scipp.github.io/scippneutron)
  scipp: {sc.__version__} (https://scipp.github.io)'''
    mantid_version = _mantid_version()
    if mantid_version:
        msg += f'\n  Mantid: {mantid_version} (https://www.mantidproject.org)'
    get_logger().info(msg)


_INSTRUMENTS = [
    'amor',
    'beer',
    'bifrost',
    'cspec',
    'dream',
    'estia',
    'freia',
    'heimdal',
    'loki',
    'magic',
    'miracles',
    'nmx',
    'odin',
    'skadi',
    'trex',
    'v20',
    'vespa',
]


def _deduce_instrument_name(f: Any) -> Optional[str]:
    # Assumes package name: ess.<instrument>[.subpackage]
    package = inspect.getmodule(f).__package__
    components = package.split('.', 2)
    try:
        if components[0] == 'ess':
            candidate = components[1]
            if candidate in _INSTRUMENTS:
                return candidate
    except IndexError:
        pass
    return None


def _function_name(f: Callable) -> str:
    if hasattr(f, '__module__'):
        return f'{f.__module__}.{f.__name__}'
    return f.__name__


def _make_stream_handler(
    level: Union[str, int], show_thread: bool, show_process: bool
) -> logging.StreamHandler:
    handler = logging.StreamHandler()
    handler.setLevel(level)
    handler.setFormatter(Formatter(show_thread, show_process))
    return handler


def _make_file_handler(
    filename: Union[str, PathLike],
    level: Union[str, int],
    show_thread: bool,
    show_process: bool,
) -> logging.FileHandler:
    handler = logging.FileHandler(filename, mode='w')
    handler.setLevel(level)
    handler.setFormatter(Formatter(show_thread, show_process))
    return handler


def _make_handlers(
    filename: Optional[Union[str, PathLike]],
    file_level: Union[str, int],
    stream_level: Union[str, int],
    widget_level: Union[str, int],
    show_thread: bool,
    show_process: bool,
) -> List[logging.Handler]:
    handlers = [_make_stream_handler(stream_level, show_thread, show_process)]
    if filename is not None:
        handlers.append(
            _make_file_handler(filename, file_level, show_thread, show_process)
        )
    if running_in_jupyter():
        handlers.append(sc.logging.make_widget_handler())
    return handlers


def _configure_logger(
    logger: logging.Logger, handlers: List[logging.Handler], level: Union[str, int]
):
    for handler in handlers:
        logger.addHandler(handler)
    logger.setLevel(level)


def _configure_mantid_logging(level: str):
    try:
        from mantid.utils.logging import log_to_python

        log_to_python(level)
    except ImportError:
        pass


def _base_level(levels: List[Union[str, int]]) -> int:
    return min(
        (
            logging.getLevelName(level) if isinstance(level, str) else level
            for level in levels
        )
    )


def _mantid_version() -> Optional[str]:
    try:
        import mantid

        return mantid.__version__
    except ImportError:
        return None



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

# flake8: noqa
import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

from . import beam_center_finder, common, conversions, i_of_q, normalization, sans2d

providers = conversions.providers + i_of_q.providers + normalization.providers
"""
List of providers for setting up a Sciline pipeline.

This provides a default workflow, including a beam-center estimation based on a
center-of-mass approach. Providers for loadings files are not included. Combine with
the providers for a specific instrument, such as :py:data:`esssans.sans2d.providers`
to setup a complete workflow.
"""
# Default to fast but potentially inaccurate beam center finder
providers.append(beam_center_finder.beam_center_from_center_of_mass)

del importlib



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
_version = '1'

__all__ = ['get_path']


def _make_pooch():
    import pooch

    return pooch.create(
        path=pooch.os_cache('esssans'),
        env='ESSSANS_DATA_DIR',
        base_url='https://public.esss.dk/groups/scipp/ess/loki/{version}/',
        version=_version,
        registry={
            'DIRECT_SANS2D_REAR_34327_4m_8mm_16Feb16.hdf5': 'md5:43f4188301d709aa49df0631d03a67cb',  # noqa: E501
            'SANS2D00063091.hdf5': 'md5:c212f1c8c68db69eae88eca90a19e7e6',
            'SANS2D00063114.hdf5': 'md5:806a5780ff02676afcea1c3d8777ee21',
            'SANS2D00063159.hdf5': 'md5:7be098bcc1f4ca73394584076a99146d',
        },
    )


_pooch = _make_pooch()


def get_path(name: str) -> str:
    """
    Return the path to a data file.

    This function only works with example data and cannot handle
    paths to custom files.
    """
    return _pooch.fetch(name)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import numpy as np
import scipp as sc

from esssans import normalization


def test_solid_angle():
    l2 = np.arange(1.0, 11.0)
    da = sc.DataArray(
        data=sc.array(dims=['x'], values=l2, unit='counts'),
        coords={'L2': sc.array(dims=['x'], values=l2, unit='m')},
    )
    da.coords['pixel_width'] = sc.scalar(2.0)
    da.coords['pixel_height'] = sc.scalar(3.0)

    solid_angle = normalization.solid_angle_rectangular_approximation(da)

    assert sc.isclose(solid_angle[0].data, solid_angle[1].data * 4).value
    assert sc.isclose(solid_angle[0].data, solid_angle[-1].data * 100).value
    da.coords['pixel_width'] *= 2
    assert sc.allclose(
        solid_angle.data * 2,
        normalization.solid_angle_rectangular_approximation(da).data,
    )
    da.coords['pixel_width'] *= 0.5
    da.coords['pixel_height'] *= 3
    assert sc.allclose(
        solid_angle.data * 3,
        normalization.solid_angle_rectangular_approximation(da).data,
    )


def test_solid_angle_keeps_relevant_masks():
    l2 = np.arange(1.0, 11.0)
    tof = sc.arange('tof', 101, unit='us')
    da = sc.DataArray(
        data=sc.array(
            dims=['spectrum', 'tof'], values=np.random.random([10, 100]), unit='counts'
        ),
        coords={'L2': sc.array(dims=['spectrum'], values=l2, unit='m'), 'tof': tof},
        masks={
            'mask1': sc.array(dims=['spectrum'], values=l2 > 5),
            'mask2': sc.array(dims=['tof'], values=tof < sc.scalar(50, unit='us')),
        },
    )
    da.coords['pixel_width'] = sc.scalar(2.0)
    da.coords['pixel_height'] = sc.scalar(3.0)

    solid_angle = normalization.solid_angle_rectangular_approximation(data=da)

    assert 'mask1' in solid_angle.masks
    assert 'mask2' not in solid_angle.masks


def test_transmission_fraction():
    N = 100
    wavelength = sc.linspace(
        dim='wavelength', start=2.0, stop=16.0, num=N + 1, unit='angstrom'
    )
    sample_incident_monitor = sc.DataArray(
        data=sc.array(
            dims=['wavelength'], values=100.0 * np.random.random(N), unit='counts'
        ),
        coords={'wavelength': wavelength},
    )
    sample_transmission_monitor = sc.DataArray(
        data=sc.array(
            dims=['wavelength'], values=50.0 * np.random.random(N), unit='counts'
        ),
        coords={'wavelength': wavelength},
    )

    direct_incident_monitor = sc.DataArray(
        data=sc.array(
            dims=['wavelength'], values=100.0 * np.random.random(N), unit='counts'
        ),
        coords={'wavelength': wavelength},
    )
    direct_transmission_monitor = sc.DataArray(
        data=sc.array(
            dims=['wavelength'], values=80.0 * np.random.random(N), unit='counts'
        ),
        coords={'wavelength': wavelength},
    )

    trans_frac = normalization.transmission_fraction(
        sample_incident_monitor=sample_incident_monitor,
        sample_transmission_monitor=sample_transmission_monitor,
        direct_incident_monitor=direct_incident_monitor,
        direct_transmission_monitor=direct_transmission_monitor,
    )

    # If counts on data transmission monitor have increased, it means less neutrons
    # have been absorbed and transmission fraction should increase.
    # - data run: incident: 100 -> transmission: 75
    # - direct run: incident: 100 -> transmission: 80
    assert sc.allclose(
        (trans_frac * sc.scalar(1.5)).data,
        normalization.transmission_fraction(
            sample_incident_monitor=sample_incident_monitor,
            sample_transmission_monitor=sample_transmission_monitor * sc.scalar(1.5),
            direct_incident_monitor=direct_incident_monitor,
            direct_transmission_monitor=direct_transmission_monitor,
        ).data,
    )

    # If counts on direct transmission monitor are higher, it means that many more
    # neutrons are absorbed when the sample is in the path of the beam, and therefore
    # the transmission fraction should decrease.
    # - data run: incident: 100 -> transmission: 50
    # - direct run: incident: 100 -> transmission: 90
    assert sc.allclose(
        (trans_frac / sc.scalar(9 / 8)).data,
        normalization.transmission_fraction(
            sample_incident_monitor=sample_incident_monitor,
            sample_transmission_monitor=sample_transmission_monitor,
            direct_incident_monitor=direct_incident_monitor,
            direct_transmission_monitor=direct_transmission_monitor * sc.scalar(9 / 8),
        ).data,
    )

    # If counts on direct incident monitor are higher, but counts on direct transmission
    # monitor are the same, it means that the relative difference between incident and
    # transmission has increased for the direct run, but not for the data run.
    # This would be the case where neutron beam flux was higher during the direct run.
    # This implies that that the transmission fraction is higher than in our vanilla
    # run.
    # - data run: incident: 100 -> transmission: 50
    # - direct run: incident: 110 -> transmission: 80
    assert sc.allclose(
        (trans_frac * sc.scalar(1.1)).data,
        normalization.transmission_fraction(
            sample_incident_monitor=sample_incident_monitor,
            sample_transmission_monitor=sample_transmission_monitor,
            direct_incident_monitor=direct_incident_monitor * sc.scalar(1.1),
            direct_transmission_monitor=direct_transmission_monitor,
        ).data,
    )

    # If counts on data incident monitor are higher, but counts on data transmission
    # monitor are the same, it means that more neutrons were absorbed in this run,
    # and then transmission fraction decreases.
    # - data run: incident: 110 -> transmission: 50
    # - direct run: incident: 100 -> transmission: 80
    assert sc.allclose(
        (trans_frac / sc.scalar(1.1)).data,
        normalization.transmission_fraction(
            sample_incident_monitor=sample_incident_monitor * sc.scalar(1.1),
            sample_transmission_monitor=sample_transmission_monitor,
            direct_incident_monitor=direct_incident_monitor,
            direct_transmission_monitor=direct_transmission_monitor,
        ).data,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""Tools for handling statistical uncertainties."""

from typing import Dict, TypeVar, Union, overload

import scipp as sc

T = TypeVar("T", bound=Union[sc.Variable, sc.DataArray])


@overload
def broadcast_with_upper_bound_variances(
    data: sc.Variable, sizes: Dict[str, int]
) -> sc.Variable:
    pass


@overload
def broadcast_with_upper_bound_variances(
    data: sc.DataArray, sizes: Dict[str, int]
) -> sc.DataArray:
    pass


def broadcast_with_upper_bound_variances(
    data: Union[sc.Variable, sc.DataArray], sizes: Dict[str, int]
) -> Union[sc.Variable, sc.DataArray]:
    if _no_variance_broadcast(data, sizes):
        return data
    size = 1
    for dim, dim_size in sizes.items():
        if dim not in data.dims:
            size *= dim_size
    data = data.copy()
    data.variances *= size
    return data.broadcast(sizes={**sizes, **data.sizes}).copy()


def drop_variances_if_broadcast(
    data: Union[sc.Variable, sc.DataArray], sizes: Dict[str, int]
) -> Union[sc.Variable, sc.DataArray]:
    if _no_variance_broadcast(data, sizes):
        return data
    return sc.values(data)


def _no_variance_broadcast(
    data: Union[sc.Variable, sc.DataArray], sizes: Dict[str, int]
) -> bool:
    return (data.variances is None) or all(
        data.sizes.get(dim) == size for dim, size in sizes.items()
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""
Loading and masking specific to the ISIS Sans2d instrument and files stored in Scipp's
HDF5 format.
"""
from typing import Optional

import scipp as sc

from .common import gravity_vector
from .types import (
    DetectorEdgeMask,
    DirectBeam,
    DirectBeamFilename,
    Filename,
    MaskedData,
    MonitorType,
    NeXusMonitorName,
    RawData,
    RawMonitor,
    RunType,
    SampleHolderMask,
    SampleRun,
)


def pooch_load(filename: Filename[RunType]) -> RawData[RunType]:
    from .data import get_path

    da = sc.io.load_hdf5(filename=get_path(filename))
    if 'gravity' not in da.coords:
        da.coords["gravity"] = gravity_vector()
    return RawData[RunType](da)


def pooch_load_direct_beam(filename: DirectBeamFilename) -> DirectBeam:
    from .data import get_path

    return DirectBeam(sc.io.load_hdf5(filename=get_path(filename)))


def get_monitor(
    da: RawData[RunType], nexus_name: NeXusMonitorName[MonitorType]
) -> RawMonitor[RunType, MonitorType]:
    # See https://github.com/scipp/sciline/issues/52 why copy needed
    mon = da.attrs[nexus_name].value.copy()
    return RawMonitor[RunType, MonitorType](mon)


def detector_edge_mask(sample: RawData[SampleRun]) -> DetectorEdgeMask:
    mask_edges = (
        sc.abs(sample.coords['position'].fields.x) > sc.scalar(0.48, unit='m')
    ) | (sc.abs(sample.coords['position'].fields.y) > sc.scalar(0.45, unit='m'))
    return DetectorEdgeMask(mask_edges)


def sample_holder_mask(sample: RawData[SampleRun]) -> SampleHolderMask:
    summed = sample.sum('tof')
    holder_mask = (
        (summed.data < sc.scalar(100, unit='counts'))
        & (sample.coords['position'].fields.x > sc.scalar(0, unit='m'))
        & (sample.coords['position'].fields.x < sc.scalar(0.42, unit='m'))
        & (sample.coords['position'].fields.y < sc.scalar(0.05, unit='m'))
        & (sample.coords['position'].fields.y > sc.scalar(-0.15, unit='m'))
    )
    return SampleHolderMask(holder_mask)


def mask_detectors(
    da: RawData[RunType],
    edge_mask: Optional[DetectorEdgeMask],
    holder_mask: Optional[SampleHolderMask],
) -> MaskedData[RunType]:
    """Apply pixel-specific masks to raw data.

    Parameters
    ----------
    da:
        Raw data.
    edge_mask:
        Mask for detector edges.
    holder_mask:
        Mask for sample holder.
    """
    da = da.copy(deep=False)
    if edge_mask is not None:
        da.masks['edges'] = edge_mask
    if holder_mask is not None:
        da.masks['holder_mask'] = holder_mask
    return MaskedData[RunType](da)


providers = [
    pooch_load_direct_beam,
    pooch_load,
    get_monitor,
    detector_edge_mask,
    sample_holder_mask,
    mask_detectors,
]
"""
Providers for loading and masking Sans2d data.

These are meant for complementing the top-level :py:data:`esssans.providers` list.
"""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Dict, Optional, Union

import scipp as sc
from scipp.scipy.interpolate import interp1d

from .common import mask_range
from .logging import get_logger
from .types import (
    BackgroundRun,
    BackgroundSubtractedIofQ,
    CleanDirectBeam,
    CleanMonitor,
    CleanQ,
    CleanSummedQ,
    DirectBeam,
    IofQ,
    IofQPart,
    MonitorType,
    NonBackgroundWavelengthRange,
    QBins,
    RunType,
    SampleRun,
    UncertaintyBroadcastMode,
    WavelengthBands,
    WavelengthBins,
    WavelengthMonitor,
)
from .uncertainty import broadcast_with_upper_bound_variances


def preprocess_monitor_data(
    monitor: WavelengthMonitor[RunType, MonitorType],
    wavelength_bins: WavelengthBins,
    non_background_range: Optional[NonBackgroundWavelengthRange],
    uncertainties: UncertaintyBroadcastMode,
) -> CleanMonitor[RunType, MonitorType]:
    """
    Prepare monitor data for computing the transmission fraction.
    The input data are first converted to wavelength (if needed).
    If a ``non_background_range`` is provided, it defines the region where data is
    considered not to be background, and regions outside are background. A mean
    background level will be computed from the background and will be subtracted from
    the non-background counts.
    Finally, if wavelength bins are provided, the data is rebinned to match the
    requested binning.

    Parameters
    ----------
    monitor:
        The monitor to be pre-processed.
    wavelength_bins:
        The binning in wavelength to use for the rebinning.
    non_background_range:
        The range of wavelengths that defines the data which does not constitute
        background. Everything outside this range is treated as background counts.
    uncertainties:
        The mode for broadcasting uncertainties. See
        :py:class:`UncertaintyBroadcastMode` for details.

    Returns
    -------
    :
        The input monitors converted to wavelength, cleaned of background counts, and
        rebinned to the requested wavelength binning.
    """
    background = None
    if non_background_range is not None:
        mask = sc.DataArray(
            data=sc.array(dims=[non_background_range.dim], values=[True]),
            coords={non_background_range.dim: non_background_range},
        )
        background = mask_range(monitor, mask=mask).mean()

    if monitor.bins is not None:
        monitor = monitor.hist(wavelength=wavelength_bins)
    else:
        monitor = monitor.rebin(wavelength=wavelength_bins)

    if background is not None:
        if uncertainties == UncertaintyBroadcastMode.drop:
            monitor -= sc.values(background)
        elif uncertainties == UncertaintyBroadcastMode.upper_bound:
            monitor -= broadcast_with_upper_bound_variances(
                background, sizes=monitor.sizes
            )
        else:
            monitor -= background
    return CleanMonitor(monitor)


def resample_direct_beam(
    direct_beam: DirectBeam, wavelength_bins: WavelengthBins
) -> CleanDirectBeam:
    """
    If the wavelength binning of the direct beam function does not match the requested
    ``wavelength_bins``, perform a 1d interpolation of the function onto the bins.

    Parameters
    ----------
    direct_beam:
        The DataArray containing the direct beam function (it should have a dimension
        of wavelength).
    wavelength_bins:
        The binning in wavelength that the direct beam function should be resampled to.

    Returns
    -------
    :
        The direct beam function resampled to the requested resolution.
    """
    if sc.identical(direct_beam.coords['wavelength'], wavelength_bins):
        return direct_beam
    func = interp1d(sc.values(direct_beam), 'wavelength')
    direct_beam = func(wavelength_bins, midpoints=True)
    logger = get_logger('sans')
    logger.warning(
        'An interpolation was performed on the direct_beam function. '
        'The variances in the direct_beam function have been dropped.'
    )
    return CleanDirectBeam(direct_beam)


def merge_spectra(
    data: CleanQ[RunType, IofQPart],
    q_bins: QBins,
    wavelength_bands: WavelengthBands,
) -> CleanSummedQ[RunType, IofQPart]:
    """
    Merges all spectra:

    * In the case of event data, events in all bins are concatenated
    * In the case of dense data, counts in all spectra are summed

    Parameters
    ----------
    data:
        A DataArray containing the data that is to be converted to Q.
    q_bins:
        The binning in Q to be used.
    wavelength_bands:
        Defines bands in wavelength that can be used to separate different wavelength
        ranges that contribute to different regions in Q space. Note that this needs to
        be defined, so if all wavelengths should be used, this should simply be a start
        and end edges that encompass the entire wavelength range.

    Returns
    -------
    :
        The input data converted to Q and then summed over all detector pixels.
    """
    if data.bins is not None:
        out = _events_merge_spectra(
            data_q=data, q_bins=q_bins, wavelength_bands=wavelength_bands
        )
    else:
        out = _dense_merge_spectra(
            data_q=data, q_bins=q_bins, wavelength_bands=wavelength_bands
        )
    if (wavelength_bands is not None) and (wavelength_bands.sizes['wavelength'] == 2):
        out = out['wavelength', 0]
    return CleanSummedQ[RunType, IofQPart](out)


def _to_q_bins(q_bins: Union[int, sc.Variable]) -> Dict[str, Union[int, sc.Variable]]:
    """
    If the input bins are an integer, convert them to a dictionary that can be used
    to bin a DataArray.
    """
    if isinstance(q_bins, int):
        return {'Q': q_bins}
    return {q_bins.dim: q_bins}


def _events_merge_spectra(
    data_q: sc.DataArray,
    q_bins: Union[int, sc.Variable],
    wavelength_bands: Optional[sc.Variable] = None,
) -> sc.DataArray:
    """
    Merge spectra of event data
    """
    q_all_pixels = data_q.bins.concat(set(data_q.dims) - {'Q'})
    edges = _to_q_bins(q_bins)
    if wavelength_bands is not None:
        edges[wavelength_bands.dim] = wavelength_bands
    return q_all_pixels.bin(**edges)


def _dense_merge_spectra(
    data_q: sc.DataArray,
    q_bins: Union[int, sc.Variable],
    wavelength_bands: Optional[sc.Variable] = None,
) -> sc.DataArray:
    """
    Merge spectra of dense data
    """
    bands = []
    sum_dims = set(data_q.dims) - {'Q'}
    edges = _to_q_bins(q_bins)
    if wavelength_bands is None:
        return data_q.hist(**edges).sum(sum_dims)
    for i in range(wavelength_bands.sizes['wavelength'] - 1):
        band = data_q['wavelength', wavelength_bands[i] : wavelength_bands[i + 1]]
        bands.append(band.hist(**edges).sum(sum_dims))
    q_summed = sc.concat(bands, 'wavelength')
    return q_summed


def subtract_background(
    sample: IofQ[SampleRun], background: IofQ[BackgroundRun]
) -> BackgroundSubtractedIofQ:
    if sample.bins is not None:
        sample = sample.bins.sum()
    if background.bins is not None:
        background = background.bins.sum()
    return BackgroundSubtractedIofQ(sample - background)


providers = [
    preprocess_monitor_data,
    resample_direct_beam,
    merge_spectra,
    subtract_background,
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Optional

import scipp as sc
import scippneutron as scn
from scipp.core import concepts

from .types import (
    CalibratedMaskedData,
    Clean,
    CleanDirectBeam,
    CleanMonitor,
    CleanSummedQ,
    Denominator,
    DirectRun,
    Incident,
    IofQ,
    NormWavelengthTerm,
    Numerator,
    RunType,
    SampleRun,
    SolidAngle,
    Transmission,
    TransmissionFraction,
    UncertaintyBroadcastMode,
)
from .uncertainty import (
    broadcast_with_upper_bound_variances,
    drop_variances_if_broadcast,
)


def solid_angle_rectangular_approximation(
    data: CalibratedMaskedData[RunType],
) -> SolidAngle[RunType]:
    """
    Solid angle computed from rectangular pixels with a 'width' and a 'height'.

    Note that this is an approximation which is only valid for small angles
    between the line of sight and the rectangle normal.

    Parameters
    ----------
    data:
        The DataArray that contains the positions for the detector pixels and the
        sample, as well as `pixel_width` and `pixel_height` as coordinates.

    Returns
    -------
    :
        The solid angle of the detector pixels, as viewed from the sample position.
        Any coords and masks that have a dimension common to the dimensions of the
        position coordinate are retained in the output.
    """
    pixel_width = data.coords['pixel_width']
    pixel_height = data.coords['pixel_height']
    L2 = scn.L2(data)
    omega = (pixel_width * pixel_height) / (L2 * L2)
    dims = set(data.dims) - set(omega.dims)
    return SolidAngle[RunType](
        concepts.rewrap_reduced_data(prototype=data, data=omega, dim=dims)
    )


def transmission_fraction(
    sample_incident_monitor: CleanMonitor[SampleRun, Incident],
    sample_transmission_monitor: CleanMonitor[SampleRun, Transmission],
    direct_incident_monitor: CleanMonitor[DirectRun, Incident],
    direct_transmission_monitor: CleanMonitor[DirectRun, Transmission],
) -> TransmissionFraction:
    """
    Approximation based on equations in
    [CalculateTransmission](https://docs.mantidproject.org/v4.0.0/algorithms/CalculateTransmission-v1.html)
    documentation:
    ``(Sample_T_monitor / Direct_T_monitor) * (Direct_I_monitor / Sample_I_monitor)``

    This is equivalent to ``mantid.CalculateTransmission`` without fitting.
    Inputs should be wavelength-dependent.

    TODO: It seems we are always multiplying this by data_monitors['incident'] to
    compute the normalization term. We could consider just returning
    ``(Sample_T_monitor / Direct_T_monitor) * Direct_I_monitor``

    Parameters
    ----------
    data_monitors:
        The data arrays for the incident and transmission monitors for the measurement
        run (monitor data should depend on wavelength).
    direct_monitors:
        The data arrays for the incident and transmission monitors for the direct
        run (monitor data should depend on wavelength).

    Returns
    -------
    :
        The transmission fraction computed from the monitor counts.
    """  # noqa: E501
    frac = (sample_transmission_monitor / direct_transmission_monitor) * (
        direct_incident_monitor / sample_incident_monitor
    )
    return TransmissionFraction(frac)


_broadcasters = {
    UncertaintyBroadcastMode.drop: drop_variances_if_broadcast,
    UncertaintyBroadcastMode.upper_bound: broadcast_with_upper_bound_variances,
    UncertaintyBroadcastMode.fail: lambda x, sizes: x,
}


def iofq_norm_wavelength_term(
    data_transmission_monitor: CleanMonitor[RunType, Transmission],
    direct_incident_monitor: CleanMonitor[DirectRun, Incident],
    direct_transmission_monitor: CleanMonitor[DirectRun, Transmission],
    direct_beam: Optional[CleanDirectBeam],
    uncertainties: UncertaintyBroadcastMode,
) -> NormWavelengthTerm[RunType]:
    """
    Compute the wavelength-dependen contribution to the denominator term for the I(Q) normalization.

    This is basically:
    ``direct_beam * data_transmission_monitor * direct_incident_monitor / direct_transmission_monitor``
    If the direct beam is not supplied, it is assumed to be 1.

    Because the multiplication between the wavelength dependent terms (monitor counts)
    and the pixel dependent term (solid angle) consists of a broadcast operation which
    would introduce correlations, variances are dropped or replaced by an upper-bound
    estimation, depending on the configured mode.

    Parameters
    ----------
    data_transmission_monitor:
        The transmission monitor counts from the measurement run (depends on
        wavelength).
    direct_incident_monitor:
        The incident monitor counts from the direct run (depends on wavelength).
    direct_transmission_monitor:
        The transmission monitor counts from the direct run (depends on wavelength).
    direct_beam:
        The DataArray containing the direct beam function (depends on wavelength).
    uncertainties:
        The mode for broadcasting uncertainties. See
        :py:class:`UncertaintyBroadcastMode` for details.

    Returns
    -------
    :
        Wavelength-dependent term for the denominator of the SANS I(Q) normalization.
        Used by :py:func:`iofq_denominator`.
    """  # noqa: E501
    denominator = (
        data_transmission_monitor
        * direct_incident_monitor
        / direct_transmission_monitor
    )

    # We need to remove the variances because the broadcasting operation between
    # solid_angle (pixel-dependent) and monitors (wavelength-dependent) will fail.
    # The direct beam may also be pixel-dependent. In this case we need to drop
    # variances of the other term already before multiplying by solid_angle.
    if direct_beam is not None:
        broadcast = _broadcasters[uncertainties]
        denominator = direct_beam * broadcast(denominator, sizes=direct_beam.sizes)
    # Convert wavelength coordinate to midpoints for future histogramming
    # if wavelength_to_midpoints:
    denominator.coords['wavelength'] = sc.midpoints(denominator.coords['wavelength'])
    return NormWavelengthTerm[RunType](denominator)


def iofq_denominator(
    wavelength_term: NormWavelengthTerm[RunType],
    solid_angle: SolidAngle[RunType],
    uncertainties: UncertaintyBroadcastMode,
) -> Clean[RunType, Denominator]:
    """
    Compute the denominator term for the I(Q) normalization.

    This is basically:
    ``solid_angle * direct_beam * data_transmission_monitor * direct_incident_monitor / direct_transmission_monitor``
    The `wavelength_term` included all but the `solid_angle` and is computed by
    :py:func:`iofq_norm_wavelength_term`.

    Because the multiplication between the wavelength dependent terms (monitor counts)
    and the pixel dependent term (solid angle) consists of a broadcast operation which
    would introduce correlations, variances are dropped or replaced by an upper-bound
    estimation, depending on the configured mode.


    Parameters
    ----------
    solid_angle:
        The solid angle of the detector pixels, as viewed from the sample position.
    wavelength_term:
        The term that depends on wavelength, computed by :py:func:`iofq_norm_wavelength_term`.
    uncertainties:
        The mode for broadcasting uncertainties. See
        :py:class:`UncertaintyBroadcastMode` for details.

    Returns
    -------
    :
        The denominator for the SANS I(Q) normalization.
    """  # noqa: E501
    broadcast = _broadcasters[uncertainties]
    denominator = solid_angle * broadcast(wavelength_term, sizes=solid_angle.sizes)
    return Clean[RunType, Denominator](denominator)


def normalize(
    numerator: CleanSummedQ[RunType, Numerator],
    denominator: CleanSummedQ[RunType, Denominator],
) -> IofQ[RunType]:
    """
    Perform normalization of counts as a function of Q.
    If the numerator contains events, we use the sc.lookup function to perform the
    division.

    Parameters
    ----------
    numerator:
        The data whose counts will be divided by the denominator. This can either be
        event or dense (histogrammed) data.
    denominator:
        The divisor for the normalization operation. This cannot be event data, it must
        contain histogrammed data.

    Returns
    -------
    :
        The input data normalized by the supplied denominator.
    """
    if denominator.variances is not None and numerator.bins is not None:
        # Event-mode normalization is not correct of norm-term has variances.
        # See https://doi.org/10.3233/JNR-220049 for context.
        numerator = numerator.hist()
    if numerator.bins is not None:
        da = numerator.bins / sc.lookup(func=denominator, dim='Q')
    else:
        da = numerator / denominator
    return IofQ[RunType](da)


providers = [
    transmission_fraction,
    iofq_norm_wavelength_term,
    iofq_denominator,
    normalize,
    solid_angle_rectangular_approximation,
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Dict, List, NewType, Optional, Union

import numpy as np
import sciline
import scipp as sc
from scipp.core import concepts

from .conversions import (
    ElasticCoordTransformGraph,
    calibrate_positions,
    detector_to_wavelength,
    mask_wavelength,
    to_Q,
)
from .i_of_q import merge_spectra
from .logging import get_logger
from .normalization import (
    iofq_denominator,
    normalize,
    solid_angle_rectangular_approximation,
)
from .types import (
    BeamCenter,
    IofQ,
    MaskedData,
    NormWavelengthTerm,
    QBins,
    SampleRun,
    UncertaintyBroadcastMode,
    WavelengthBands,
    WavelengthBins,
)


def _xy_extrema(pos: sc.Variable) -> sc.Variable:
    x_min = pos.fields.x.min()
    x_max = pos.fields.x.max()
    y_min = pos.fields.y.min()
    y_max = pos.fields.y.max()
    return sc.concat([x_min, x_max, y_min, y_max], dim='extremes')


def beam_center_from_center_of_mass(
    data: MaskedData[SampleRun],
    graph: ElasticCoordTransformGraph,
) -> BeamCenter:
    """
    Estimate the beam center via the center-of-mass of the data counts.

    The center-of-mass is simply the weighted mean of the positions.
    Areas with low counts are excluded from the center of mass calculation, as they
    typically fall into asymmetric regions of the detector panel and would thus lead
    to a biased result. The beam is assumed to be roughly aligned with the Z axis.
    The returned beam center is the component normal to the beam direction, projected
    onto the X-Y plane.

    Parameters
    ----------
    data:
        The data to find the beam center of.
    graph:
        Coordinate transformation graph for elastic SANS.

    Returns
    -------
    :
        The beam center position as a vector.
    """

    summed = data.sum(list(set(data.dims) - set(data.coords['position'].dims)))
    v = sc.values(summed)
    mask = concepts.irreducible_mask(summed, dim=None)
    pos = data.coords['position']
    extrema = _xy_extrema(pos[~mask])
    # Mean including existing masks
    cutoff = 0.1 * v.mean().data
    low_counts = v.data < cutoff
    # Increase cutoff until we no longer include pixels at the X/Y min/max.
    # This would be simpler if the logical panel shape was reflected in the
    # dims of the input data, instead of having a flat list of pixels.
    while sc.any(_xy_extrema(pos[~(mask | low_counts)]) == extrema):
        cutoff *= 2.0
        low_counts = v.data < cutoff
    # See scipp/scipp#3271, the following lines are a workaround
    select = ~(low_counts | mask)
    v = v.data[select]
    pos = pos[select]
    com = sc.sum(pos * v) / v.sum()
    # We compute the shift between the incident beam direction and the center-of-mass
    incident_beam = data.transform_coords('incident_beam', graph=graph).coords[
        'incident_beam'
    ]
    n_beam = incident_beam / sc.norm(incident_beam)
    com_shift = com - sc.dot(com, n_beam) * n_beam
    xy = [com_shift.fields.x.value, com_shift.fields.y.value]
    return _offsets_to_vector(data=data, xy=xy, graph=graph)


def _offsets_to_vector(data: sc.DataArray, xy: List[float], graph: dict) -> sc.Variable:
    """
    Convert x,y offsets inside the plane normal to the beam to a vector in absolute
    coordinates.
    """
    u = data.coords['position'].unit
    # Get two vectors that define the plane normal to the beam
    coords = data.transform_coords(
        ['cyl_x_unit_vector', 'cyl_y_unit_vector'], graph=graph
    ).coords
    center = xy[0] * coords['cyl_x_unit_vector'] + xy[1] * coords['cyl_y_unit_vector']
    center.unit = u
    return center


def _iofq_in_quadrants(
    xy: List[float],
    data: sc.DataArray,
    norm: sc.DataArray,
    graph: dict,
    q_bins: Union[int, sc.Variable],
    wavelength_range: sc.Variable,
) -> Dict[str, sc.DataArray]:
    """
    Compute the intensity as a function of Q inside 4 quadrants in Phi.

    Parameters
    ----------
    xy:
        The x,y offsets in the plane normal to the beam.
    data:
        The sample data.
    norm:
        The denominator data for normalization.
    graph:
        Coordinate transformation graph.
    q_bins:
        Bin edges for Q.
    wavelength_range:
        The wavelength range to use for computing the intensity as a function of Q.

    Returns
    -------
    :
        A dictionary containing the intensity as a function of Q in each quadrant.
        The quadrants are named 'south-west', 'south-east', 'north-east', and
        'north-west'.
    """
    pi = sc.constants.pi.value
    phi = data.transform_coords(
        'phi', graph=graph, keep_intermediate=False, keep_inputs=False
    ).coords['phi']
    phi_bins = sc.linspace('phi', -pi, pi, 5, unit='rad')
    quadrants = ['south-west', 'south-east', 'north-east', 'north-west']

    providers = [
        to_Q,
        merge_spectra,
        normalize,
        iofq_denominator,
        mask_wavelength,
        detector_to_wavelength,
        calibrate_positions,
        solid_angle_rectangular_approximation,
    ]
    params = {}
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound
    params[WavelengthBands] = wavelength_range
    params[QBins] = q_bins
    params[ElasticCoordTransformGraph] = graph
    params[BeamCenter] = _offsets_to_vector(data=data, xy=xy, graph=graph)

    out = {}
    for i, quad in enumerate(quadrants):
        # Select pixels based on phi
        sel = (phi >= phi_bins[i]) & (phi < phi_bins[i + 1])
        params[MaskedData[SampleRun]] = data[sel]
        params[NormWavelengthTerm[SampleRun]] = (
            norm if norm.dims == ('wavelength',) else norm[sel]
        )
        pipeline = sciline.Pipeline(providers, params=params)
        out[quad] = pipeline.compute(IofQ[SampleRun])
    return out


def _cost(xy: List[float], *args) -> float:
    """
    Cost function for determining how close the :math:`I(Q)` curves are in all four
    quadrants. The cost is defined as

    .. math::

       \\text{cost} = \\frac{\\sum_{Q}\\sum_{i=1}^{i=4} \\overline{I}(Q)\\left(I(Q)_{i} - \\overline{I}(Q)\\right)^2}{\\sum_{Q}\\overline{I}(Q)} ~,

    where :math:`i` represents the 4 quadrants and :math:`\\overline{I}(Q)` is the mean
    intensity of the 4 quadrants as a function of :math:`Q`. This is basically a
    weighted mean of the square of the differences between the :math:`I(Q)` curves in
    the 4 quadrants with respect to the mean, and where the weights are
    :math:`\\overline{I}(Q)`.
    We use a weighted mean, as opposed to relative (percentage) differences to give
    less importance to regions with low statistics which are potentially noisy and
    would contribute significantly to the computed cost.

    Parameters
    ----------
    xy:
        The x,y offsets in the plane normal to the beam.
    *args:
        Arguments passed to :func:`iofq_in_quadrants`.

    Returns
    -------
    :
        The sum of the residuals for :math:`I(Q)` in the 4 quadrants, with respect to
        the mean :math:`I(Q)` in all quadrants.

    Notes
    -----
    Mantid uses a different cost function. They compute the horizontal (Left - Right)
    and the vertical (Top - Bottom) costs, and require both to be below the tolerance.
    The costs are defined as

    .. math::

       \\text{cost} = \\sum_{Q} \\left(I(Q)_{\\text{L,T}} - I(Q)_{\\text{R,B}}\\right)^2 ~.

    Using absolute differences instead of a weighted mean is similar to our cost
    function in the way that it would give a lot of weight to even a small difference
    in a high-intensity region. However, it also means that an absolute difference of
    e.g. 2 in a high-intensity region would be weighted the same as a difference of 2
    in a low-intensity region.
    It is also not documented why two separate costs are computed, instead of a single
    one. The Mantid implementation is available
    `here <https://github.com/mantidproject/mantid/blob/main/Framework/PythonInterface/plugins/algorithms/WorkflowAlgorithms/SANS/SANSBeamCentreFinder.py`_.
    """  # noqa: E501
    iofq = _iofq_in_quadrants(xy, *args)
    all_q = sc.concat([sc.values(da) for da in iofq.values()], dim='quadrant')
    ref = all_q.mean('quadrant')
    c = (all_q - ref) ** 2
    out = (sc.sum(ref * c) / sc.sum(ref)).value
    logger = get_logger('sans')
    logger.info(f'Beam center finder: x={xy[0]}, y={xy[1]}, cost={out}')
    if not np.isfinite(out):
        raise ValueError(
            'Non-finite value computed in cost. This is likely due to a division by '
            'zero. Try restricting your Q range, or increasing the size of your Q bins '
            'to improve statistics in the denominator.'
        )
    return out


BeamCenterFinderQBins = NewType('BeamCenterFinderQBins', sc.Variable)
"""Q binning used for the beam center finder"""

BeamCenterFinderTolerance = NewType('BeamCenterFinderTolerance', float)
"""Tolerance used for the beam center finder"""

BeamCenterFinderMinimizer = NewType('BeamCenterFinderMinimizer', str)
"""Minimizer used for the beam center finder"""


def beam_center_from_iofq(
    data: MaskedData[SampleRun],
    graph: ElasticCoordTransformGraph,
    wavelength_bins: WavelengthBins,
    norm: NormWavelengthTerm[SampleRun],
    q_bins: BeamCenterFinderQBins,
    minimizer: Optional[BeamCenterFinderMinimizer],
    tolerance: Optional[BeamCenterFinderTolerance],
) -> BeamCenter:
    """
    Find the beam center of a SANS scattering pattern using an I(Q) calculation.

    Description of the procedure:

    #. obtain an initial guess by computing the center-of-mass of the pixels,
       weighted by the counts on each pixel
    #. from that initial guess, divide the panel into 4 quadrants
    #. compute :math:`I(Q)` inside each quadrant and compute the residual difference
       between all 4 quadrants
    #. iteratively move the centre position and repeat 2. and 3. until all 4
       :math:`I(Q)` curves lie on top of each other

    Parameters
    ----------
    data:
        The DataArray containing the detector data.
    graph:
        Coordinate transformation graph for elastic SANS.
    wavelength_bins:
        The binning in the wavelength dimension to be used.
    q_bins:
        The binning in the Q dimension to be used.
    minimizer:
        The Scipy minimizer method to use (see the
        `Scipy docs <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_
        for details).
    tolerance:
        Tolerance for termination (see the
        `Scipy docs <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_
        for details).

    Returns
    -------
    :
        The beam center position as a vector.

    Notes
    -----
    We record here the thought process we went through during the writing of this
    algorithm. This information is important for understanding why the beam center
    finding is implemented the way it is, and should be considered carefully before
    making changes to the logic of the algorithm.

    **Use a + cut, not an X cut**

    The first idea for implementing the beam center finder was to cut the detector
    panel into 4 wedges using a cross (X) shape. This is what Mantid does, and seemed
    natural, because the offsets when searching for the beam center would be applied
    along the horizontal and vertical directions.
    This worked well on square detector panels (like the SANS2D detector), but on
    rectangular detectors, the north and south wedges ended up holding many less pixels
    than the east and west panels.
    More pixels means more contributions to a particular :math:`Q` bin, and comparing
    the :math:`I(Q)` curves in the 4 wedges was thus not possible.
    We therefore divided the detector panel into 4 quadrants using a ``+`` cut instead.
    Note that since we are looking at an isotropic scattering pattern, the shape of the
    cut (and the number of quadrants) should not matter for the resulting shapes of the
    :math:`I(Q)` curves.

    **Normalization inside the 4 quadrants**

    The first attempt at implementing the beam center finder was to only compute the
    raw counts as a function of $Q$ for the sample run, and not compute any
    normalization term.
    The idea was that even though this would change the shape of the :math:`I(Q)` curve,
    because we were looking at isotropic scattering, it would change the shape of the
    curve isotropically, thus still allowing us to find the center when the curves in
    all 4 quadrants overlap.
    The motivation for this was to save computational cost.

    After discovering the issue that using a ``X`` shaped cut for dividing the detector
    panel would yield different contributions to :math:`I(Q)` in the different wedges,
    we concluded that some normalization was necessary.
    The first version was to simply sum the counts in each quadrant and use this to
    normalize the counts for each intensity curve.

    This was, however, not sufficient in cases where masks are applied to the detector
    pixels. It is indeed very common to mask broken pixels, as well as the region of
    the detector where the sample holder is casting a shadow.
    Such a sample holder will not appear in all 4 quadrants, and because it spans a
    range of scattering (:math:`2{\\theta}`) angles, it spans a range of :math:`Q` bins.

    All this means that we in fact need to perform a reduction as close as possible to
    the full :math:`I(Q)` reduction in each of the 4 quadrants to achieve a reliable
    result.
    We write 'as close as possible' because In the full :math:`I(Q)` reduction, there
    is a term :math:`D({\\lambda})` in the normalization called the 'direct beam' which
    gives the efficiency of the detectors as a function of wavelength.
    Because finding the beam center is required to compute the direct beam in the first
    place, we do not include this term in the computation of :math:`I(Q)` for finding
    the beam center. This changes the shape of the :math:`I(Q)` curve, but since it
    changes it in the same manner for all :math:`{\\phi}` angles, this does not affect the
    results for finding the beam center.

    This is what is now implemented in this version of the algorithm.
    """  # noqa: E501
    from scipy.optimize import minimize

    logger = get_logger('sans')

    logger.info(f'Requested minimizer: {minimizer}')
    logger.info(f'Requested tolerance: {tolerance}')
    minimizer = minimizer or 'Nelder-Mead'
    tolerance = tolerance or 0.1
    logger.info(f'Using minimizer: {minimizer}')
    logger.info(f'Using tolerance: {tolerance}')

    # Use center of mass to get initial guess for beam center
    com_shift = beam_center_from_center_of_mass(data, graph)
    logger.info(f'Initial guess for beam center: {com_shift}')

    wavelength_range = sc.concat(
        [wavelength_bins.min(), wavelength_bins.max()], dim='wavelength'
    )

    coords = data.transform_coords(
        ['cylindrical_x', 'cylindrical_y'], graph=graph
    ).coords
    bounds = [
        (coords['cylindrical_x'].min().value, coords['cylindrical_x'].max().value),
        (coords['cylindrical_y'].min().value, coords['cylindrical_y'].max().value),
    ]

    # Refine using Scipy optimize
    res = minimize(
        _cost,
        x0=[com_shift.fields.x.value, com_shift.fields.y.value],
        args=(data, norm, graph, q_bins, wavelength_range),
        bounds=bounds,
        method=minimizer,
        tol=tolerance,
    )

    center = _offsets_to_vector(data=data, xy=res.x, graph=graph)
    logger.info(f'Final beam center value: {center}')
    logger.info(f'Beam center finder minimizer info: {res}')
    return center



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import scipp as sc
from scipp.testing import assert_identical

from esssans.uncertainty import broadcast_with_upper_bound_variances


def test_broadcast_returns_original_if_no_new_dims():
    var = sc.ones(dims=['x', 'y'], shape=[2, 3], with_variances=True)
    assert broadcast_with_upper_bound_variances(var, {'x': 2}) is var
    assert broadcast_with_upper_bound_variances(var, {'y': 3}) is var


def test_broadcast_returns_original_if_no_variances():
    var = sc.ones(dims=['x'], shape=[2], with_variances=False)
    assert broadcast_with_upper_bound_variances(var, {'y': 3}) is var


def test_broadcast_scales_variances_by_new_subspace_volume():
    x = sc.linspace('x', 0.0, 1.0, 2)
    y = sc.linspace('y', 0.0, 2.0, 3)
    values = x * y
    var = values.copy()
    var.variances = var.values
    expected = sc.ones(dims=['z'], shape=[1]) * values
    expected.variances = 1 * expected.values
    assert_identical(broadcast_with_upper_bound_variances(var, {'z': 1}), expected)
    expected = sc.ones(dims=['z'], shape=[2]) * values
    expected.variances = 2 * expected.values
    assert_identical(broadcast_with_upper_bound_variances(var, {'z': 2}), expected)
    expected = sc.ones(dims=['z'], shape=[2]) * values
    expected.variances = 2 * expected.values
    assert_identical(
        broadcast_with_upper_bound_variances(var, {'y': 3, 'z': 2}),
        expected.transpose(['y', 'z', 'x']),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc

from esssans.common import mask_range


def test_mask_range_dense_data():
    x = sc.arange('x', 5.0, unit='m')
    da = sc.DataArray(data=x, coords={'x': x})
    mask = sc.DataArray(
        data=sc.array(dims=['x'], values=[True]),
        coords={
            'x': sc.array(dims=['x'], values=[1.9, 3.001], unit='m'),
        },
    )
    masked = mask_range(da, mask=mask, name='mymask')
    assert masked.masks['mymask'].values.tolist() == [False, False, True, True, False]


def test_mask_range_dense_data_bin_edges():
    x = sc.arange('x', 6.0, unit='m')
    da = sc.DataArray(data=x[:-1], coords={'x': x})
    mask = sc.DataArray(
        data=sc.array(dims=['x'], values=[True]),
        coords={
            'x': sc.array(dims=['x'], values=[1.9, 3.001], unit='m'),
        },
    )
    masked = mask_range(da, mask=mask, name='mymask')
    assert masked.masks['mymask'].values.tolist() == [False, True, True, True, False]


def test_mask_range_dense_data_two_ranges():
    x = sc.arange('x', 11.0, unit='m')
    da = sc.DataArray(data=x[:-1], coords={'x': x})
    mask = sc.DataArray(
        data=sc.array(dims=['x'], values=[True, False, True]),
        coords={
            'x': sc.array(dims=['x'], values=[1.9, 3.001, 6.5, 8.8], unit='m'),
        },
    )
    masked = mask_range(da, mask=mask, name='mymask')
    assert masked.masks['mymask'].values.tolist() == [
        False,
        True,
        True,
        True,
        False,
        False,
        True,
        True,
        True,
        False,
    ]


def test_mask_range_binned_data_no_prior_binning():
    x = sc.arange('x', 5.0, unit='m')
    da = sc.DataArray(data=x, coords={'x': x, 'y': x + x.max()})
    binned = da.bin(y=2)
    mask = sc.DataArray(
        data=sc.array(dims=['x'], values=[True]),
        coords={
            'x': sc.array(dims=['x'], values=[1.9, 3.001], unit='m'),
        },
    )
    masked = mask_range(binned, mask=mask, name='mymask')
    assert masked.masks['mymask'].values.tolist() == [True]
    assert sc.identical(masked.coords['x'], mask.coords['x'])


def test_mask_range_binned_data_has_prior_single_bin():
    x = sc.arange('x', 5.0, unit='m')
    da = sc.DataArray(data=x, coords={'x': x, 'y': x + x.max()})
    binned = da.bin(x=1)
    mask = sc.DataArray(
        data=sc.array(dims=['x'], values=[True]),
        coords={
            'x': sc.array(dims=['x'], values=[1.5, 3.001], unit='m'),
        },
    )
    masked = mask_range(binned, mask=mask, name='mymask')
    assert masked.masks['mymask'].values.tolist() == [False, True, False]
    assert np.allclose(masked.coords['x'].values, [0.0, 1.5, 3.001, 4.0])


def test_mask_range_binned_data_has_prior_multiple_bins():
    x = sc.arange('x', 5.0, unit='m')
    da = sc.DataArray(data=x, coords={'x': x, 'y': x + x.max()})
    binned = da.bin(x=2)
    mask = sc.DataArray(
        data=sc.array(dims=['x'], values=[True]),
        coords={
            'x': sc.array(dims=['x'], values=[1.5, 3.001], unit='m'),
        },
    )
    masked = mask_range(binned, mask=mask, name='mymask')
    assert masked.masks['mymask'].values.tolist() == [False, True, True, False]
    assert np.allclose(masked.coords['x'].values, [0.0, 1.5, 2.0, 3.001, 4.0])


def test_mask_range_binned_data_has_already_same_edge_as_mask():
    x = sc.arange('x', 5.0, unit='m')
    da = sc.DataArray(data=x, coords={'x': x, 'y': x + x.max()})
    binned = da.bin(x=2)
    mask = sc.DataArray(
        data=sc.array(dims=['x'], values=[True]),
        coords={
            'x': sc.array(
                dims=['x'], values=[binned.coords['x'][1].value, 3.001], unit='m'
            ),
        },
    )
    masked = mask_range(binned, mask=mask, name='mymask')
    assert masked.masks['mymask'].values.tolist() == [False, True, False]
    assert np.allclose(masked.coords['x'].values, [0.0, 2.0, 3.001, 4.0])


def test_mask_range_with_midpoints_coord_raises():
    x = sc.arange('x', 5.0, unit='m')
    da = sc.DataArray(data=x, coords={'x': x})
    mask = sc.DataArray(
        data=sc.array(dims=['x'], values=[True, False]),
        coords={
            'x': sc.array(dims=['x'], values=[1.9, 3.001], unit='m'),
        },
    )
    with pytest.raises(
        sc.DimensionError, match='Coordinate x must be bin-edges to mask a range'
    ):
        _ = mask_range(da, mask=mask, name='mymask')


def test_mask_range_on_dense_data_with_two_dimensional_coord_raises():
    xy = sc.arange('x', 20.0, unit='m').fold(dim='x', sizes={'y': 4, 'x': 5})
    da = sc.DataArray(data=xy, coords={'x': xy})
    mask = sc.DataArray(
        data=sc.array(dims=['x'], values=[True]),
        coords={
            'x': sc.array(dims=['x'], values=[1.9, 3.001], unit='m'),
        },
    )
    with pytest.raises(
        sc.DimensionError,
        match='Cannot mask range on data with multi-dimensional coordinate',
    ):
        _ = mask_range(da, mask=mask, name='mymask')


def test_mask_range_on_binned_data_with_two_dimensional_coord_raises():
    x = sc.arange('x', 10.0, unit='m')
    da = sc.DataArray(data=x, coords={'x': x, 'y': x * 1.5})
    binned = da.bin(x=2, y=2)
    binned.coords['x'] = sc.broadcast(binned.coords['x'], sizes={'x': 3, 'y': 2})
    mask = sc.DataArray(
        data=sc.array(dims=['x'], values=[True]),
        coords={
            'x': sc.array(dims=['x'], values=[1.9, 3.001], unit='m'),
        },
    )
    with pytest.raises(
        sc.DimensionError,
        match='Cannot mask range on data with multi-dimensional coordinate',
    ):
        _ = mask_range(binned, mask=mask, name='mymask')


def test_mask_range_on_data_with_existing_mask_of_same_name_raises():
    x = sc.arange('x', 10.0, unit='m')
    da = sc.DataArray(
        data=x,
        coords={'x': x, 'y': x * 1.5},
        masks={'mymask': x > sc.scalar(5.0, unit='m')},
    )
    mask = sc.DataArray(
        data=sc.array(dims=['x'], values=[True]),
        coords={
            'x': sc.array(dims=['x'], values=[1.9, 3.001], unit='m'),
        },
    )
    with pytest.raises(
        ValueError,
        match='Mask mymask already exists in data array and would be overwritten',
    ):
        _ = mask_range(da, mask=mask, name='mymask')



import sys
from argparse import ArgumentParser

import tomli

parser = ArgumentParser()
parser.add_argument(
    "--nightly",
    default="",
    help="List of dependencies to install from main branch for nightly tests, "
    "separated by commas.",
)
args = parser.parse_args()

header = "# Generated by 'tox -e deps', DO NOT EDIT MANUALLY!'\n"

with open("../pyproject.toml", "rb") as toml_file:
    pyproject = tomli.load(toml_file)
    dependencies = pyproject["project"].get("dependencies")
    if not dependencies:
        raise RuntimeError("No dependencies found in pyproject.toml")
    dependencies = [dep.strip().strip('"') for dep in dependencies]

with open("base.in", "w") as f:
    f.write(header)
    f.write("\n".join(dependencies))


def as_nightly(repo: str) -> str:
    if "/" in repo:
        org, repo = repo.split("/")
    else:
        org = "scipp"
    if repo == "scipp":
        version = f"cp{sys.version_info.major}{sys.version_info.minor}"
        base = "https://github.com/scipp/scipp/releases/download/nightly/scipp-nightly"
        suffix = "manylinux_2_17_x86_64.manylinux2014_x86_64.whl"
        return "-".join([base, version, version, suffix])
    return f"{repo} @ git+https://github.com/{org}/{repo}@main"


nightly = args.nightly.split(",") if args.nightly else []
dependencies = [dep for dep in dependencies if not dep.startswith(tuple(nightly))]
dependencies += [as_nightly(arg) for arg in nightly]

with open("nightly.in", "w") as f:
    f.write(header)
    f.write("\n".join(dependencies))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import NewType, Optional

import scipp as sc
from scipp.constants import h, m_n
from scippneutron._utils import elem_unit
from scippneutron.conversion.graph import beamline, tof

from .common import mask_range
from .types import (
    BeamCenter,
    CalibratedMaskedData,
    Clean,
    CleanMasked,
    CleanQ,
    CorrectForGravity,
    IofQPart,
    MaskedData,
    MonitorType,
    Numerator,
    RawMonitor,
    RunType,
    WavelengthMask,
    WavelengthMonitor,
)


def cyl_x_unit_vector(gravity: sc.Variable, incident_beam: sc.Variable) -> sc.Variable:
    """
    Compute the horizontal unit vector in the plane normal to the incident beam
    direction. Note that it is assumed here that the incident beam is perpendicular to
    the gravity vector.
    """
    v_x = sc.cross(incident_beam, gravity)
    return v_x / sc.norm(v_x)


def cyl_y_unit_vector(gravity: sc.Variable) -> sc.Variable:
    """
    Compute the vertical unit vector in the plane normal to the incident beam
    direction. Note that it is assumed here that the incident beam is perpendicular to
    the gravity vector.
    """
    v_y = -gravity
    return v_y / sc.norm(v_y)


def cylindrical_x(
    cyl_x_unit_vector: sc.Variable, scattered_beam: sc.Variable
) -> sc.Variable:
    """
    Compute the horizontal x coordinate perpendicular to the incident beam direction.
    Note that it is assumed here that the incident beam is perpendicular to the gravity
    vector.
    """
    return sc.dot(scattered_beam, cyl_x_unit_vector)


def cylindrical_y(
    cyl_y_unit_vector: sc.Variable, scattered_beam: sc.Variable
) -> sc.Variable:
    """
    Compute the vertical y coordinate perpendicular to the incident beam direction.
    Note that it is assumed here that the incident beam is perpendicular to the gravity
    vector.
    """
    return sc.dot(scattered_beam, cyl_y_unit_vector)


def two_theta(
    incident_beam: sc.Variable,
    scattered_beam: sc.Variable,
    wavelength: sc.Variable,
    gravity: sc.Variable,
) -> sc.Variable:
    """
    Compute the scattering angle from the incident and scattered beam vectors, taking
    into account the effects of gravity.
    Note that it is assumed here that the incident beam is perpendicular to the gravity
    vector.
    """
    grav = sc.norm(gravity)
    L2 = sc.norm(scattered_beam)

    x_term = cylindrical_x(cyl_x_unit_vector(gravity, incident_beam), scattered_beam)
    x_term *= x_term

    y_term = sc.to_unit(wavelength, elem_unit(L2), copy=True)
    y_term *= y_term
    drop = L2**2
    drop *= grav * (m_n**2 / (2 * h**2))
    # Optimization when handling either the dense or the event coord of binned data:
    # - For the event coord, both operands have same dims, and we can multiply in place
    # - For the dense coord, we need to broadcast using non in-place operation
    if set(drop.dims).issubset(set(y_term.dims)):
        y_term *= drop
    else:
        y_term = drop * y_term
    y_term += cylindrical_y(cyl_y_unit_vector(gravity), scattered_beam)
    y_term *= y_term

    if set(x_term.dims).issubset(set(y_term.dims)):
        y_term += x_term
    else:
        y_term = y_term + x_term
    out = sc.sqrt(y_term, out=y_term)
    out /= L2
    out = sc.asin(out, out=out)
    return out


def phi(cylindrical_x: sc.Variable, cylindrical_y: sc.Variable) -> sc.Variable:
    """
    Compute the cylindrial phi angle around the incident beam. Note that it is assumed
    here that the incident beam is perpendicular to the gravity vector.
    """
    return sc.atan2(y=cylindrical_y, x=cylindrical_x)


ElasticCoordTransformGraph = NewType('ElasticCoordTransformGraph', dict)
MonitorCoordTransformGraph = NewType('MonitorCoordTransformGraph', dict)


def sans_elastic(gravity: Optional[CorrectForGravity]) -> ElasticCoordTransformGraph:
    """
    Generate a coordinate transformation graph for SANS elastic scattering.
    By default, the effects of gravity on the neutron flight paths are not included.

    :param gravity: Take into account the bending of the neutron flight paths from the
        Earth's gravitational field if ``True``.
    """
    graph = {**beamline.beamline(scatter=True), **tof.elastic_Q('tof')}
    if gravity:
        graph['two_theta'] = two_theta
    graph['cyl_x_unit_vector'] = cyl_x_unit_vector
    graph['cyl_y_unit_vector'] = cyl_y_unit_vector
    graph['cylindrical_x'] = cylindrical_x
    graph['cylindrical_y'] = cylindrical_y
    graph['phi'] = phi
    return ElasticCoordTransformGraph(graph)


def sans_monitor() -> MonitorCoordTransformGraph:
    """
    Generate a coordinate transformation graph for SANS monitor (no scattering).
    """
    return MonitorCoordTransformGraph(
        {**beamline.beamline(scatter=False), **tof.elastic_wavelength('tof')}
    )


def monitor_to_wavelength(
    monitor: RawMonitor[RunType, MonitorType], graph: MonitorCoordTransformGraph
) -> WavelengthMonitor[RunType, MonitorType]:
    return WavelengthMonitor[RunType, MonitorType](
        monitor.transform_coords('wavelength', graph=graph)
    )


def calibrate_positions(
    detector: MaskedData[RunType], beam_center: BeamCenter
) -> CalibratedMaskedData[RunType]:
    """
    Calibrate pixel positions.

    Currently the only applied calibration is the beam-center offset.
    """
    detector = detector.copy(deep=False)
    detector.coords['position'] = detector.coords['position'] - beam_center
    return detector


# TODO This demonstrates a problem: Transforming to wavelength should be possible
# for RawData, MaskedData, ... no reason to restrict necessarily.
# Would we be fine with just choosing on option, or will this get in the way for users?
def detector_to_wavelength(
    detector: CalibratedMaskedData[RunType],
    graph: ElasticCoordTransformGraph,
) -> Clean[RunType, Numerator]:
    return Clean[RunType, Numerator](
        detector.transform_coords('wavelength', graph=graph)
    )


def mask_wavelength(
    da: Clean[RunType, IofQPart], mask: Optional[WavelengthMask]
) -> CleanMasked[RunType, IofQPart]:
    if mask is not None:
        # If we have binned data and the wavelength coord is multi-dimensional, we need
        # to make a single wavelength bin before we can mask the range.
        if da.bins is not None:
            dim = mask.dim
            if (dim in da.bins.coords) and (dim in da.coords):
                da = da.bin({dim: 1})
        da = mask_range(da, mask=mask)
    return CleanMasked[RunType, IofQPart](da)


def to_Q(
    data: CleanMasked[RunType, IofQPart], graph: ElasticCoordTransformGraph
) -> CleanQ[RunType, IofQPart]:
    """
    Convert a data array from wavelength to Q.
    """
    return CleanQ[RunType, IofQPart](data.transform_coords('Q', graph=graph))


providers = [
    sans_elastic,
    sans_monitor,
    calibrate_positions,
    monitor_to_wavelength,
    detector_to_wavelength,
    mask_wavelength,
    to_Q,
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""
This modules defines the domain types uses in esssans.

The domain types are used to define parameters and to request results from a Sciline
pipeline."""
from enum import Enum
from typing import NewType, TypeVar

import sciline
import scipp as sc

# 1  TypeVars used to parametrize the generic parts of the workflow

# 1.1  Run types
BackgroundRun = NewType('BackgroundRun', int)
"""Background run"""
DirectRun = NewType('DirectRun', int)
"""Direct run"""
SampleRun = NewType('SampleRun', int)
"""Sample run"""
RunType = TypeVar('RunType', BackgroundRun, DirectRun, SampleRun)
"""TypeVar used for specifying BackgroundRun, DirectRun or SampleRun"""

# 1.2  Monitor types
Incident = NewType('Incident', int)
"""Incident monitor"""
Transmission = NewType('Transmission', int)
"""Transmission monitor"""
MonitorType = TypeVar('MonitorType', Incident, Transmission)
"""TypeVar used for specifying Incident or Transmission monitor type"""

# 1.3  Numerator and denominator of IofQ
Numerator = NewType('Numerator', sc.DataArray)
"""Numerator of IofQ"""
Denominator = NewType('Denominator', sc.DataArray)
"""Denominator of IofQ"""
IofQPart = TypeVar('IofQPart', Numerator, Denominator)
"""TypeVar used for specifying Numerator or Denominator of IofQ"""

# 2  Workflow parameters

UncertaintyBroadcastMode = Enum(
    'UncertaintyBroadcastMode', ['drop', 'upper_bound', 'fail']
)
"""
Mode for broadcasting uncertainties.

See https://doi.org/10.3233/JNR-220049 for context.
"""

WavelengthBins = NewType('WavelengthBins', sc.Variable)
"""Wavelength binning"""

WavelengthBands = NewType('WavelengthBands', sc.Variable)
"""Wavelength bands. Typically a single band, set to first and last value of
WavelengthBins."""

QBins = NewType('QBins', sc.Variable)
"""Q binning"""

NonBackgroundWavelengthRange = NewType('NonBackgroundWavelengthRange', sc.Variable)
"""Range of wavelengths that are not considered background in the monitor"""

DirectBeamFilename = NewType('DirectBeamFilename', str)
"""Filename of direct beam correction"""

BeamCenter = NewType('BeamCenter', sc.Variable)
"""Beam center, may be set directly or computed using beam-center finder"""

WavelengthMask = NewType('WavelengthMask', sc.DataArray)
"""Optional wavelength mask"""

CorrectForGravity = NewType('CorrectForGravity', bool)
"""Whether to correct for gravity when computing wavelength and Q."""


class NeXusMonitorName(sciline.Scope[MonitorType, str], str):
    """Name of Incident|Transmission monitor in NeXus file"""


class Filename(sciline.Scope[RunType, str], str):
    """Filename of BackgroundRun|DirectRun|SampleRun"""


# 3  Workflow (intermediate) results

DetectorEdgeMask = NewType('DetectorEdgeMask', sc.Variable)
"""Detector edge mask"""
SampleHolderMask = NewType('SampleHolderMask', sc.Variable)
"""Sample holder mask"""

DirectBeam = NewType('DirectBeam', sc.DataArray)
"""Direct beam"""

TransmissionFraction = NewType('TransmissionFraction', sc.DataArray)
"""
Transmission fraction for inspection purposes.

The IofQ computation does not use this, it uses the monitors directly.
"""


CleanDirectBeam = NewType('CleanDirectBeam', sc.DataArray)
"""Direct beam after resampling to required wavelength bins"""


class SolidAngle(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Solid angle of detector pixels seen from sample position"""


class RawData(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Raw data"""


class MaskedData(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Raw data with pixel-specific masks applied"""


class CalibratedMaskedData(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Raw data with pixel-specific masks applied and calibrated pixel positions"""


class NormWavelengthTerm(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Normalization term (numerator) for IofQ before scaling with solid-angle."""


class Clean(sciline.ScopeTwoParams[RunType, IofQPart, sc.DataArray], sc.DataArray):
    """
    Prerequisite for IofQ numerator or denominator.

    This can either be the sample or background counts, converted to wavelength,
    or the respective normalization terms computed from the respective solid angle,
    direct beam, and monitors.
    """


class CleanMasked(
    sciline.ScopeTwoParams[RunType, IofQPart, sc.DataArray], sc.DataArray
):
    """Result of applying wavelength masking to :py:class:`Clean`"""


class CleanQ(sciline.ScopeTwoParams[RunType, IofQPart, sc.DataArray], sc.DataArray):
    """Result of converting :py:class:`CleanMasked` to Q"""


class CleanSummedQ(
    sciline.ScopeTwoParams[RunType, IofQPart, sc.DataArray], sc.DataArray
):
    """Result of histogramming/binning :py:class:`CleanQ` over all pixels into Q bins"""


class IofQ(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """I(Q)"""


BackgroundSubtractedIofQ = NewType('BackgroundSubtractedIofQ', sc.DataArray)
"""I(Q) with background (given by I(Q) of the background run) subtracted"""


class RawMonitor(
    sciline.ScopeTwoParams[RunType, MonitorType, sc.DataArray], sc.DataArray
):
    """Raw monitor data"""


class WavelengthMonitor(
    sciline.ScopeTwoParams[RunType, MonitorType, sc.DataArray], sc.DataArray
):
    """Monitor data converted to wavelength"""


class CleanMonitor(
    sciline.ScopeTwoParams[RunType, MonitorType, sc.DataArray], sc.DataArray
):
    """Monitor data cleaned of background counts"""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import uuid
from typing import Optional

import numpy as np
import scipp as sc
from scipp.constants import g


def gravity_vector() -> sc.Variable:
    """
    Return a vector of 3 components, defining the magnitude and direction of the Earth's
    gravitational field.
    """
    return sc.vector(value=[0, -1, 0]) * g


def mask_range(
    da: sc.DataArray, mask: sc.DataArray, name: Optional[str] = None
) -> sc.DataArray:
    """
    Mask a range on a data array.
    The provided edges are used to define the ranges to be masked.

    Parameters
    ----------
    da:
        The data array to be masked.
    mask:
        A data array defining the mask to be applied. Only one-dimensional masks are
        supported. The data array should contain a bin-edge coordinate which represents
        the edges of the ranges to be masked. The values of the data array represent the
        mask values (``True`` or ``False``) inside each range defined by the coordinate.
    name:
        The name of the mask to be applied. If not provided, a random name will be used.

    Returns
    -------
    :
        A copy of the input data array with the mask applied.
    """
    if name is None:
        name = uuid.uuid4().hex
    if name in da.masks:
        raise ValueError(
            f'Mask {name} already exists in data array and would be overwritten.'
        )
    dim = mask.dim
    edges = mask.coords[dim]
    if not mask.coords.is_edges(dim):
        raise sc.DimensionError(
            f'Coordinate {dim} must be bin-edges to mask a range, found midpoints.'
        )
    if (dim in da.coords) and (da.coords[dim].ndim > 1):
        raise sc.DimensionError(
            'Cannot mask range on data with multi-dimensional coordinate. '
            f'Found dimensions {da.coords[dim].dims} for coordinate {dim}.'
        )

    lu = sc.DataArray(data=mask.data, coords={dim: edges})
    if da.bins is not None:
        if dim in da.coords:
            new_bins = sc.array(
                dims=[dim],
                values=np.union1d(edges.values, da.coords[dim].values),
                unit=edges.unit,
            )
            out = da.bin({dim: new_bins})
            out.masks[name] = sc.lookup(lu, dim)[sc.midpoints(new_bins, dim=dim)]
        else:
            out = da.bin({dim: edges})
            out.masks[name] = mask.data
    else:
        out = da.copy(deep=False)
        mask_values = sc.lookup(lu, dim)[da.coords[dim]]
        if da.coords.is_edges(dim):
            out.masks[name] = mask_values[dim, 1:] | mask_values[dim, :-1]
        else:
            out.masks[name] = mask_values
    return out
